#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['snmp',
 'snmp.manager',
 'snmp.message',
 'snmp.security',
 'snmp.security.usm',
 'snmp.security.usm.priv',
 'snmp.security.usm.priv.openssl',
 'snmp.security.usm.priv.pycryptodome',
 'snmp.transport']

package_data = \
{'': ['*'],
 'snmp': ['cffi/*', 'v3/*'],
 'snmp.transport': ['generic/*', 'posix/*']}

install_requires = \
['pycryptodome>=3.4']

setup(name='snmp',
      version='0.8.0',
      description='A user-friendly SNMP library',
      author=None,
      author_email='"Charles C. D. Tolley" <charlescdtolley@protonmail.com>',
      url=None,
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      python_requires='>=3.6',
     )
